#!/usr/bin/env python
"""
.. codeauthor:: D Litwin, G Tucker

.. sectionauthor:: D Litwin, G Tucker
"""

from .dupuit_percolator import GroundwaterDupuitPercolator

__all__ = ["GroundwaterDupuitPercolator"]
